@@ -1,7 +1,7 @@
 /*
- *  Copyright (C) 2014-2016 Savoir-faire Linux Inc.
+ *  Copyright (C) 2014-2017 Savoir-faire Linux Inc.
  *  Author(s) : Adrien Béraud <adrien.beraud@savoirfairelinux.com>
- *              Simon Désaulniers <sim.desaulniers@gmail.com>
+ *              Simon Désaulniers <simon.desaulniers@savoirfairelinux.com>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -14,7 +14,7 @@
  *  GNU General Public License for more details.
  *
  *  You should have received a copy of the GNU General Public License
- *  along with this program. If not, see <http://www.gnu.org/licenses/>.
+ *  along with this program. If not, see <https://www.gnu.org/licenses/>.
  */
 
 
@@ -25,7 +25,19 @@
 
 #include <msgpack.hpp>
 
+#ifndef _WIN32
+#include <arpa/inet.h>
+#include <unistd.h>
+#else
+#include <ws2tcpip.h>
+#include <io.h>
+#endif
+#include <fcntl.h>
+
+#include <cstring>
+
 namespace dht {
+namespace net {
 
 const std::string DhtProtocolException::GET_NO_INFOHASH {"Get_values with no info_hash"};
 const std::string DhtProtocolException::LISTEN_NO_INFOHASH {"Listen with no info_hash"};
@@ -40,15 +52,15 @@ constexpr std::chrono::seconds NetworkEngine::RX_MAX_PACKET_TIME;
 constexpr std::chrono::seconds NetworkEngine::RX_TIMEOUT;
 
 const std::string NetworkEngine::my_v {"RNG1"};
-const constexpr uint16_t NetworkEngine::TransId::INVALID;
+const constexpr uint16_t TransId::INVALID;
 std::mt19937 NetworkEngine::rd_device {dht::crypto::random_device{}()};
 
-const NetworkEngine::TransPrefix NetworkEngine::TransPrefix::PING = {"pn"};
-const NetworkEngine::TransPrefix NetworkEngine::TransPrefix::FIND_NODE = {"fn"};
-const NetworkEngine::TransPrefix NetworkEngine::TransPrefix::GET_VALUES = {"gt"};
-const NetworkEngine::TransPrefix NetworkEngine::TransPrefix::ANNOUNCE_VALUES = {"pt"};
-const NetworkEngine::TransPrefix NetworkEngine::TransPrefix::REFRESH = {"rf"};
-const NetworkEngine::TransPrefix NetworkEngine::TransPrefix::LISTEN = {"lt"};
+const TransPrefix TransPrefix::PING = {"pn"};
+const TransPrefix TransPrefix::FIND_NODE = {"fn"};
+const TransPrefix TransPrefix::GET_VALUES = {"gt"};
+const TransPrefix TransPrefix::ANNOUNCE_VALUES = {"pt"};
+const TransPrefix TransPrefix::REFRESH = {"rf"};
+const TransPrefix TransPrefix::LISTEN = {"lt"};
 constexpr long unsigned NetworkEngine::MAX_REQUESTS_PER_SEC;
 
 static const uint8_t v4prefix[16] = {
@@ -57,6 +69,32 @@ static const uint8_t v4prefix[16] = {
 
 constexpr unsigned SEND_NODES {8};
 
+#ifdef _WIN32
+
+static bool
+set_nonblocking(int fd, int nonblocking)
+{
+    unsigned long mode = !!nonblocking;
+    int rc = ioctlsocket(fd, FIONBIO, &mode);
+    return rc == 0;
+}
+
+extern const char *inet_ntop(int, const void *, char *, socklen_t);
+
+#else
+
+static bool
+set_nonblocking(int fd, int nonblocking)
+{
+    int rc = fcntl(fd, F_GETFL, 0);
+    if (rc < 0)
+        return false;
+    rc = fcntl(fd, F_SETFL, nonblocking?(rc | O_NONBLOCK):(rc & ~O_NONBLOCK));
+    return rc >= 0;
+}
+
+#endif
+
 enum class MessageType {
     Error = 0,
     Reply,
@@ -80,7 +118,9 @@ struct ParsedMessage {
     /* target id around which to find nodes */
     InfoHash target;
     /* transaction id */
-    NetworkEngine::TransId tid;
+    TransId tid;
+    /* tid for packets going through request socket */
+    TransId socket_id;
     /* security token */
     Blob token;
     /* the value id (announce confirmation) */
@@ -89,11 +129,11 @@ struct ParsedMessage {
     time_point created { time_point::max() };
     /* IPv4 nodes in response to a 'find' request */
     Blob nodes4_raw, nodes6_raw;
-    std::vector<std::shared_ptr<Node>> nodes4, nodes6;
+    std::vector<Sp<Node>> nodes4, nodes6;
     /* values to store or retreive request */
-    std::vector<std::shared_ptr<Value>> values;
+    std::vector<Sp<Value>> values;
     /* index for fields values */
-    std::vector<std::shared_ptr<FieldValueIndex>> fields;
+    std::vector<Sp<FieldValueIndex>> fields;
     /** When part of the message header: {index -> (total size, {})}
      *  When part of partial value data: {index -> (offset, part_data)} */
     std::map<unsigned, std::pair<unsigned, Blob>> value_parts;
@@ -120,7 +160,7 @@ struct NetworkEngine::PartialMessage {
 };
 
 std::vector<Blob>
-serializeValues(const std::vector<std::shared_ptr<Value>>& st)
+serializeValues(const std::vector<Sp<Value>>& st)
 {
     std::vector<Blob> svals;
     svals.reserve(st.size());
@@ -148,6 +188,14 @@ NetworkEngine::NetworkEngine(InfoHash& myid, NetId net, int s, int s6, Logger& l
     onGetValues(onGetValues), onListen(onListen), onAnnounce(onAnnounce), onRefresh(onRefresh), myid(myid),
     network(net), dht_socket(s), dht_socket6(s6), DHT_LOG(log), scheduler(scheduler)
 {
+    if (dht_socket >= 0) {
+        if (!set_nonblocking(dht_socket, 1))
+            throw DhtException("Can't set socket to non-blocking mode");
+    }
+    if (dht_socket6 >= 0) {
+        if (!set_nonblocking(dht_socket6, 1))
+            throw DhtException("Can't set socket to non-blocking mode");
+    }
     transaction_id = std::uniform_int_distribution<decltype(transaction_id)>{1}(rd_device);
 }
 
@@ -156,15 +204,14 @@ NetworkEngine::~NetworkEngine() {
 }
 
 void
-NetworkEngine::tellListener(std::shared_ptr<Node> node, uint16_t rid, const InfoHash& hash, want_t want,
-        const Blob& ntoken, std::vector<std::shared_ptr<Node>>&& nodes,
-        std::vector<std::shared_ptr<Node>>&& nodes6, std::vector<std::shared_ptr<Value>>&& values,
+NetworkEngine::tellListener(Sp<Node> node, uint32_t socket_id, const InfoHash& hash, want_t want,
+        const Blob& ntoken, std::vector<Sp<Node>>&& nodes,
+        std::vector<Sp<Node>>&& nodes6, std::vector<Sp<Value>>&& values,
         const Query& query)
 {
     auto nnodes = bufferNodes(node->getFamily(), hash, want, nodes, nodes6);
     try {
-        sendNodesValues(node->addr, TransId {TransPrefix::GET_VALUES, (uint16_t)rid}, nnodes.first, nnodes.second,
-                values, query, ntoken);
+        sendNodesValues(node->addr, TransId((char*)&socket_id, 4), nnodes.first, nnodes.second, values, query, ntoken);
     } catch (const std::overflow_error& e) {
         DHT_LOG.e("Can't send value: buffer not large enough !");
     }
@@ -185,11 +232,29 @@ NetworkEngine::isRunning(sa_family_t af) const
     }
 }
 
+Sp<Socket>
+NetworkEngine::openSocket(const Sp<Node>& node, TransPrefix tp, SocketCb&& cb)
+{
+    auto tid = TransId {tp, getNewTid()};
+    auto s = opened_sockets.emplace(tid, std::make_shared<Socket>(node, tid, cb));
+    if (not s.second)
+        DHT_LOG.e(node->id, "[node %s] socket (tid: %d) already opened!", node->id.toString().c_str(), tid.toInt());
+    return s.first->second;
+}
+
 void
-NetworkEngine::cancelRequest(std::shared_ptr<Request>& req)
+NetworkEngine::closeSocket(Sp<Socket> socket)
+{
+    if (socket)
+        opened_sockets.erase(socket->id);
+}
+
+void
+NetworkEngine::cancelRequest(Sp<Request>& req)
 {
     if (req) {
         req->cancel();
+        closeSocket(req->socket);
         requests.erase(req->tid);
     }
 }
@@ -209,7 +274,7 @@ NetworkEngine::connectivityChanged(sa_family_t af)
 }
 
 void
-NetworkEngine::requestStep(std::shared_ptr<Request> sreq)
+NetworkEngine::requestStep(Sp<Request> sreq)
 {
     auto& req = *sreq;
     if (not req.pending()) {
@@ -229,16 +294,26 @@ NetworkEngine::requestStep(std::shared_ptr<Request> sreq)
         req.on_expired(req, false);
     }
 
-    send((char*)req.msg.data(), req.msg.size(),
+    auto err = send((char*)req.msg.data(), req.msg.size(),
             (node.reply_time >= now - UDP_REPLY_TIME) ? 0 : MSG_CONFIRM,
             node.addr);
-    ++req.attempt_count;
-    req.last_try = now;
-    std::weak_ptr<Request> wreq = sreq;
-    scheduler.add(req.last_try + Node::MAX_RESPONSE_TIME, [this,wreq] {
-        if (auto req = wreq.lock())
-            requestStep(req);
-    });
+    if (err == ENETUNREACH  ||
+        err == EHOSTUNREACH ||
+        err == EAFNOSUPPORT)
+    {
+        node.setExpired();
+        requests.erase(req.tid);
+    } else {
+        if (err != EAGAIN) {
+            ++req.attempt_count;
+        }
+        req.last_try = now;
+        std::weak_ptr<Request> wreq = sreq;
+        scheduler.add(req.last_try + Node::MAX_RESPONSE_TIME, [this,wreq] {
+            if (auto req = wreq.lock())
+                requestStep(req);
+        });
+    }
 }
 
 /**
@@ -246,7 +321,7 @@ NetworkEngine::requestStep(std::shared_ptr<Request> sreq)
  * be made before the request expires.
  */
 void
-NetworkEngine::sendRequest(std::shared_ptr<Request>& request)
+NetworkEngine::sendRequest(Sp<Request>& request)
 {
     request->start = scheduler.time();
     auto e = requests.emplace(request->tid, request);
@@ -255,7 +330,7 @@ NetworkEngine::sendRequest(std::shared_ptr<Request>& request)
         // Try to handle this scenario as well as we can
         e.first->second->setExpired();
         e.first->second = request;
-        DHT_LOG.e(request->node->id, "Request already existed (tid: %d)!", request->tid);
+        DHT_LOG.e(request->node->id, "Request already existed (tid: %d)!", request->tid.getTid());
     }
     request->node->requested(request);
     requestStep(request);
@@ -287,32 +362,25 @@ NetworkEngine::rateLimit(const SockAddr& addr)
 bool
 NetworkEngine::isMartian(const SockAddr& addr)
 {
-    // Check that sa_family can be accessed safely
-    if (addr.second < sizeof(sockaddr_in))
+    if (addr.getPort() == 0)
         return true;
-
-    switch(addr.first.ss_family) {
+    switch(addr.getFamily()) {
     case AF_INET: {
-        sockaddr_in *sin = (sockaddr_in*)&addr.first;
-        const uint8_t *address = (const uint8_t*)&sin->sin_addr;
-        return sin->sin_port == 0 ||
-            (address[0] == 0) ||
-            (address[0] == 127) ||
-            ((address[0] & 0xE0) == 0xE0);
+        const auto& sin = addr.getIPv4();
+        const uint8_t* address = (const uint8_t*)&sin.sin_addr;
+        return (address[0] == 0) ||
+              ((address[0] & 0xE0) == 0xE0);
     }
     case AF_INET6: {
         if (addr.second < sizeof(sockaddr_in6))
             return true;
-        sockaddr_in6 *sin6 = (sockaddr_in6*)&addr.first;
-        const uint8_t *address = (const uint8_t*)&sin6->sin6_addr;
-        return sin6->sin6_port == 0 ||
-            (address[0] == 0xFF) ||
-            (address[0] == 0xFE && (address[1] & 0xC0) == 0x80) ||
-            (memcmp(address, zeroes.data(), 15) == 0 &&
-             (address[15] == 0 || address[15] == 1)) ||
-            (memcmp(address, v4prefix, 12) == 0);
+        const auto& sin6 = addr.getIPv6();
+        const uint8_t* address = (const uint8_t*)&sin6.sin6_addr;
+        return address[0] == 0xFF ||
+              (address[0] == 0xFE && (address[1] & 0xC0) == 0x80) ||
+               memcmp(address, zeroes.data(), 16) == 0 ||
+               memcmp(address, v4prefix,      12) == 0;
     }
-
     default:
         return true;
     }
@@ -321,7 +389,7 @@ NetworkEngine::isMartian(const SockAddr& addr)
 /* The internal blacklist is an LRU cache of nodes that have sent
    incorrect messages. */
 void
-NetworkEngine::blacklistNode(const std::shared_ptr<Node>& n)
+NetworkEngine::blacklistNode(const Sp<Node>& n)
 {
     n->setExpired();
     for (auto rit = requests.begin(); rit != requests.end();) {
@@ -435,19 +503,22 @@ NetworkEngine::process(std::unique_ptr<ParsedMessage>&& msg, const SockAddr& fro
     const auto& now = scheduler.time();
 
     if (msg->type == MessageType::Error or msg->type == MessageType::Reply) {
-        auto reqp = requests.find(msg->tid.getTid());
-        if (reqp == requests.end()) {
+        /* either response for a request or data for an opened socket */
+        auto req_it = requests.find(msg->tid);
+        auto rsocket_it = opened_sockets.end();
+        if (req_it == requests.end())
+            rsocket_it = opened_sockets.find(msg->tid);
+        if (req_it == requests.end() and rsocket_it == opened_sockets.end())
             throw DhtProtocolException {DhtProtocolException::UNKNOWN_TID, "Can't find transaction", msg->id};
-        }
-        auto req = reqp->second;
-        auto node = req->node;
+
+        auto req = req_it != requests.end() ? req_it->second : nullptr;
+        auto rsocket = rsocket_it != opened_sockets.end() ? rsocket_it->second : nullptr;
+
+        auto& node = req ? req->node : rsocket->node;
         if (node->id != msg->id) {
-            bool unknown_node = node->id == zeroes;
-            node = cache.getNode(msg->id, from, now, true);
-            if (unknown_node) {
-                // received reply to a message sent when we didn't know the node ID.
-                req->node = node;
-            } else {
+            if (node->id == zeroes) // received reply to a message sent when we didn't know the node ID.
+                node = cache.getNode(msg->id, from, now, true);
+            else {
                 // received reply from unexpected node
                 node->received(now, req);
                 onNewNode(node, 2);
@@ -456,14 +527,15 @@ NetworkEngine::process(std::unique_ptr<ParsedMessage>&& msg, const SockAddr& fro
             }
         } else
             node->update(from);
+
         node->received(now, req);
 
         onNewNode(node, 2);
         onReportedAddr(msg->id, msg->addr);
 
-        if (req->cancelled() or req->expired() or (req->completed() and not req->persistent)) {
+        if (req and (req->cancelled() or req->expired() or req->completed())) {
             DHT_LOG.w(node->id, "[node %s] response to expired, cancelled or completed request", node->toString().c_str());
-            requests.erase(reqp);
+            requests.erase(req_it);
             return;
         }
 
@@ -474,8 +546,8 @@ NetworkEngine::process(std::unique_ptr<ParsedMessage>&& msg, const SockAddr& fro
                 (msg->error_code == DhtProtocolException::UNAUTHORIZED and (msg->tid.matches(TransPrefix::ANNOUNCE_VALUES)
                                                                          or msg->tid.matches(TransPrefix::LISTEN)))))
             {
-                req->last_try = TIME_INVALID;
-                req->reply_time = TIME_INVALID;
+                req->last_try = time_point::min();
+                req->reply_time = time_point::min();
                 onError(req, DhtProtocolException {msg->error_code});
             } else {
                 DHT_LOG.w(msg->id, "[node %s %s] received unknown error message %u",
@@ -484,17 +556,21 @@ NetworkEngine::process(std::unique_ptr<ParsedMessage>&& msg, const SockAddr& fro
             break;
         }
         case MessageType::Reply:
-            if (msg->type == MessageType::AnnounceValue or msg->type == MessageType::Listen)
-                req->node->authSuccess();
+         if (req) { /* request reply */
+             if (msg->type == MessageType::AnnounceValue or msg->type == MessageType::Listen)
+                 req->node->authSuccess();
 
-            // erase before calling callback to make sure iterator is still valid
-            if (not req->persistent)
-                requests.erase(reqp);
-            req->reply_time = scheduler.time();
+             // erase before calling callback to make sure iterator is still valid
+             requests.erase(req_it);
+             req->reply_time = scheduler.time();
 
-            deserializeNodes(*msg);
-            req->setDone(std::move(*msg));
-            break;
+             deserializeNodes(*msg, from);
+             req->setDone(std::move(*msg));
+             break;
+         } else { /* request socket data */
+             deserializeNodes(*msg, from);
+             rsocket->on_receive(rsocket->node, std::move(*msg));
+         }
         default:
             break;
         }
@@ -548,7 +624,9 @@ NetworkEngine::process(std::unique_ptr<ParsedMessage>&& msg, const SockAddr& fro
             case MessageType::Listen: {
                 DHT_LOG.d(msg->info_hash, node->id, "[node %s] got 'listen' request for %s", node->toString().c_str(), msg->info_hash.toString().c_str());
                 ++in_stats.listen;
-                RequestAnswer answer = onListen(node, msg->info_hash, msg->token, msg->tid.getTid(), std::move(msg->query));
+                /* TODO: backward compatibility check to remove in a few versions (see ::sendListen doc) */
+                auto socket_id = msg->socket_id.getTid() ? msg->socket_id : TransId { TransPrefix::GET_VALUES, msg->tid.getTid() };
+                RequestAnswer answer = onListen(node, msg->info_hash, msg->token, socket_id.toInt(), std::move(msg->query));
                 sendListenConfirmation(from, msg->tid);
                 break;
             }
@@ -564,7 +642,7 @@ NetworkEngine::process(std::unique_ptr<ParsedMessage>&& msg, const SockAddr& fro
 }
 
 void
-packToken(msgpack::packer<msgpack::sbuffer>& pk, Blob token)
+packToken(msgpack::packer<msgpack::sbuffer>& pk, const Blob& token)
 {
     pk.pack_bin(token.size());
     pk.pack_bin_body((char*)token.data(), token.size());
@@ -586,7 +664,7 @@ int
 NetworkEngine::send(const char *buf, size_t len, int flags, const SockAddr& addr)
 {
     if (addr.second == 0)
-        return -1;
+        return EFAULT;
 
     int s;
     if (addr.getFamily() == AF_INET)
@@ -597,12 +675,17 @@ NetworkEngine::send(const char *buf, size_t len, int flags, const SockAddr& addr
         s = -1;
 
     if (s < 0)
-        return -1;
-    return sendto(s, buf, len, flags, (const sockaddr*)&addr.first, addr.second);
+        return EAFNOSUPPORT;
+    if (sendto(s, buf, len, flags, (const sockaddr*)&addr.first, addr.second) == -1) {
+        int err = errno;
+        DHT_LOG.e("Can't send message to %s: %s", addr.toString().c_str(), strerror(err));
+        return err;
+    }
+    return 0;
 }
 
-std::shared_ptr<Request>
-NetworkEngine::sendPing(std::shared_ptr<Node> node, RequestCb on_done, RequestExpiredCb on_expired) {
+Sp<Request>
+NetworkEngine::sendPing(Sp<Node> node, RequestCb&& on_done, RequestExpiredCb&& on_expired) {
     auto tid = TransId {TransPrefix::PING, getNewTid()};
     msgpack::sbuffer buffer;
     msgpack::packer<msgpack::sbuffer> pk(&buffer);
@@ -620,8 +703,8 @@ NetworkEngine::sendPing(std::shared_ptr<Node> node, RequestCb on_done, RequestEx
         pk.pack(std::string("n")); pk.pack(network);
     }
 
-    Blob b {buffer.data(), buffer.data() + buffer.size()};
-    std::shared_ptr<Request> req(new Request {tid.getTid(), node, std::move(b),
+    auto req = std::make_shared<Request>(tid, node,
+        Blob(buffer.data(), buffer.data() + buffer.size()),
         [=](const Request& req_status, ParsedMessage&&) {
             DHT_LOG.d(req_status.node->id, "[node %s] got pong !", req_status.node->toString().c_str());
             if (on_done) {
@@ -633,7 +716,7 @@ NetworkEngine::sendPing(std::shared_ptr<Node> node, RequestCb on_done, RequestEx
                 on_expired(req_status, done);
             }
         }
-    });
+    );
     sendRequest(req);
     ++out_stats.ping;
     return req;
@@ -660,9 +743,9 @@ NetworkEngine::sendPong(const SockAddr& addr, TransId tid) {
     send(buffer.data(), buffer.size(), 0, addr);
 }
 
-std::shared_ptr<Request>
-NetworkEngine::sendFindNode(std::shared_ptr<Node> n, const InfoHash& target, want_t want,
-        RequestCb on_done, RequestExpiredCb on_expired) {
+Sp<Request>
+NetworkEngine::sendFindNode(Sp<Node> n, const InfoHash& target, want_t want,
+        RequestCb&& on_done, RequestExpiredCb&& on_expired) {
     auto tid = TransId {TransPrefix::FIND_NODE, getNewTid()};
     msgpack::sbuffer buffer;
     msgpack::packer<msgpack::sbuffer> pk(&buffer);
@@ -687,8 +770,8 @@ NetworkEngine::sendFindNode(std::shared_ptr<Node> n, const InfoHash& target, wan
         pk.pack(std::string("n")); pk.pack(network);
     }
 
-    Blob b {buffer.data(), buffer.data() + buffer.size()};
-    std::shared_ptr<Request> req(new Request {tid.getTid(), n, std::move(b),
+    auto req = std::make_shared<Request>(tid, n,
+        Blob(buffer.data(), buffer.data() + buffer.size()),
         [=](const Request& req_status, ParsedMessage&& msg) { /* on done */
             if (on_done) {
                 on_done(req_status, {std::forward<ParsedMessage>(msg)});
@@ -699,16 +782,16 @@ NetworkEngine::sendFindNode(std::shared_ptr<Node> n, const InfoHash& target, wan
                 on_expired(req_status, done);
             }
         }
-    });
+    );
     sendRequest(req);
     ++out_stats.find;
     return req;
 }
 
 
-std::shared_ptr<Request>
-NetworkEngine::sendGetValues(std::shared_ptr<Node> n, const InfoHash& info_hash, const Query& query, want_t want,
-        RequestCb on_done, RequestExpiredCb on_expired) {
+Sp<Request>
+NetworkEngine::sendGetValues(Sp<Node> n, const InfoHash& info_hash, const Query& query, want_t want,
+        RequestCb&& on_done, RequestExpiredCb&& on_expired) {
     auto tid = TransId {TransPrefix::GET_VALUES, getNewTid()};
     msgpack::sbuffer buffer;
     msgpack::packer<msgpack::sbuffer> pk(&buffer);
@@ -736,8 +819,8 @@ NetworkEngine::sendGetValues(std::shared_ptr<Node> n, const InfoHash& info_hash,
         pk.pack(std::string("n")); pk.pack(network);
     }
 
-    Blob b {buffer.data(), buffer.data() + buffer.size()};
-    std::shared_ptr<Request> req(new Request {tid.getTid(), n, std::move(b),
+    auto req = std::make_shared<Request>(tid, n,
+        Blob(buffer.data(), buffer.data() + buffer.size()),
         [=](const Request& req_status, ParsedMessage&& msg) { /* on done */
             if (on_done) {
                 on_done(req_status, {std::forward<ParsedMessage>(msg)});
@@ -748,31 +831,49 @@ NetworkEngine::sendGetValues(std::shared_ptr<Node> n, const InfoHash& info_hash,
                 on_expired(req_status, done);
             }
         }
-    });
+    );
     sendRequest(req);
     ++out_stats.get;
     return req;
 }
 
+SockAddr deserializeIPv4(const uint8_t* ni) {
+    SockAddr addr;
+    auto& sin = addr.getIPv4();
+    std::fill_n((uint8_t*)&sin, sizeof(sockaddr_in), 0);
+    addr.setFamily(AF_INET);
+    memcpy(&sin.sin_addr, ni, 4);
+    memcpy(&sin.sin_port, ni + 4, 2);
+    return addr;
+}
+SockAddr deserializeIPv6(const uint8_t* ni) {
+    SockAddr addr;
+    auto& sin6 = addr.getIPv6();
+    std::fill_n((uint8_t*)&sin6, sizeof(sockaddr_in6), 0);
+    addr.setFamily(AF_INET6);
+    memcpy(&sin6.sin6_addr, ni, 16);
+    memcpy(&sin6.sin6_port, ni + 16, 2);
+    return addr;
+}
+
 void
-NetworkEngine::deserializeNodes(ParsedMessage& msg) {
+NetworkEngine::deserializeNodes(ParsedMessage& msg, const SockAddr& from) {
     if (msg.nodes4_raw.size() % NODE4_INFO_BUF_LEN != 0 || msg.nodes6_raw.size() % NODE6_INFO_BUF_LEN != 0) {
         throw DhtProtocolException {DhtProtocolException::WRONG_NODE_INFO_BUF_LEN};
     } else {
         // deserialize nodes
         const auto& now = scheduler.time();
-        SockAddr addr;
         for (unsigned i = 0; i < msg.nodes4_raw.size() / NODE4_INFO_BUF_LEN; i++) {
             uint8_t *ni = msg.nodes4_raw.data() + i * NODE4_INFO_BUF_LEN;
             const InfoHash& ni_id = *reinterpret_cast<InfoHash*>(ni);
             if (ni_id == myid)
                 continue;
-            auto sin = (sockaddr_in*)&addr.first;
-            std::fill_n((uint8_t*)sin, sizeof(sockaddr_in), 0);
-            sin->sin_family = AF_INET;
-            memcpy(&sin->sin_addr, ni + ni_id.size(), 4);
-            memcpy(&sin->sin_port, ni + ni_id.size() + 4, 2);
-            addr.second = sizeof(sockaddr_in);
+            SockAddr addr = deserializeIPv4(ni + ni_id.size());
+            if (addr.isLoopback() and from.getFamily() == AF_INET) {
+                auto port = addr.getPort();
+                addr = from;
+                addr.setPort(port);
+            }
             if (isMartian(addr) || isNodeBlacklisted(addr))
                 continue;
             msg.nodes4.emplace_back(cache.getNode(ni_id, addr, now, false));
@@ -783,12 +884,12 @@ NetworkEngine::deserializeNodes(ParsedMessage& msg) {
             const InfoHash& ni_id = *reinterpret_cast<InfoHash*>(ni);
             if (ni_id == myid)
                 continue;
-            auto sin6 = (sockaddr_in6*)&addr.first;
-            std::fill_n((uint8_t*)sin6, sizeof(sockaddr_in6), 0);
-            sin6->sin6_family = AF_INET6;
-            memcpy(&sin6->sin6_addr, ni + HASH_LEN, 16);
-            memcpy(&sin6->sin6_port, ni + HASH_LEN + 16, 2);
-            addr.second = sizeof(sockaddr_in6);
+            SockAddr addr = deserializeIPv6(ni + ni_id.size());
+            if (addr.isLoopback() and from.getFamily() == AF_INET6) {
+                auto port = addr.getPort();
+                addr = from;
+                addr.setPort(port);
+            }
             if (isMartian(addr) || isNodeBlacklisted(addr))
                 continue;
             msg.nodes6.emplace_back(cache.getNode(ni_id, addr, now, false));
@@ -798,7 +899,7 @@ NetworkEngine::deserializeNodes(ParsedMessage& msg) {
 }
 
 std::vector<Blob>
-NetworkEngine::packValueHeader(msgpack::sbuffer& buffer, const std::vector<std::shared_ptr<Value>>& st)
+NetworkEngine::packValueHeader(msgpack::sbuffer& buffer, const std::vector<Sp<Value>>& st)
 {
     auto svals = serializeValues(st);
     size_t total_size = 0;
@@ -853,7 +954,8 @@ NetworkEngine::sendValueParts(TransId tid, const std::vector<Blob>& svals, const
 
 void
 NetworkEngine::sendNodesValues(const SockAddr& addr, TransId tid, const Blob& nodes, const Blob& nodes6,
-        const std::vector<std::shared_ptr<Value>>& st, const Query& query, const Blob& token) {
+        const std::vector<Sp<Value>>& st, const Query& query, const Blob& token)
+{
     msgpack::sbuffer buffer;
     msgpack::packer<msgpack::sbuffer> pk(&buffer);
     pk.pack_map(4+(network?1:0));
@@ -909,34 +1011,32 @@ NetworkEngine::sendNodesValues(const SockAddr& addr, TransId tid, const Blob& no
 }
 
 Blob
-NetworkEngine::bufferNodes(sa_family_t af, const InfoHash& id, std::vector<std::shared_ptr<Node>>& nodes)
+NetworkEngine::bufferNodes(sa_family_t af, const InfoHash& id, std::vector<Sp<Node>>& nodes)
 {
-    std::sort(nodes.begin(), nodes.end(), [&](const std::shared_ptr<Node>& a, const std::shared_ptr<Node>& b){
+    std::sort(nodes.begin(), nodes.end(), [&](const Sp<Node>& a, const Sp<Node>& b){
         return id.xorCmp(a->id, b->id) < 0;
     });
     size_t nnode = std::min<size_t>(SEND_NODES, nodes.size());
     Blob bnodes;
     if (af == AF_INET) {
         bnodes.resize(NODE4_INFO_BUF_LEN * nnode);
-        const constexpr size_t size = HASH_LEN + sizeof(in_addr) + sizeof(in_port_t); // 26
         for (size_t i=0; i<nnode; i++) {
             const Node& n = *nodes[i];
             sockaddr_in *sin = (sockaddr_in*)&n.addr.first;
-            auto dest = bnodes.data() + size * i;
+            auto dest = bnodes.data() + NODE4_INFO_BUF_LEN * i;
             memcpy(dest, n.id.data(), HASH_LEN);
             memcpy(dest + HASH_LEN, &sin->sin_addr, sizeof(in_addr));
-            memcpy(dest + HASH_LEN + sizeof(in_addr), &sin->sin_port, 2);
+            memcpy(dest + HASH_LEN + sizeof(in_addr), &sin->sin_port, sizeof(in_port_t));
         }
     } else if (af == AF_INET6) {
         bnodes.resize(NODE6_INFO_BUF_LEN * nnode);
-        const constexpr size_t size = HASH_LEN + sizeof(in6_addr) + sizeof(in_port_t); // 38
         for (size_t i=0; i<nnode; i++) {
             const Node& n = *nodes[i];
             sockaddr_in6 *sin6 = (sockaddr_in6*)&n.addr.first;
-            auto dest = bnodes.data() + size * i;
+            auto dest = bnodes.data() + NODE6_INFO_BUF_LEN * i;
             memcpy(dest, n.id.data(), HASH_LEN);
             memcpy(dest + HASH_LEN, &sin6->sin6_addr, sizeof(in6_addr));
-            memcpy(dest + HASH_LEN + sizeof(in6_addr), &sin6->sin6_port, 2);
+            memcpy(dest + HASH_LEN + sizeof(in6_addr), &sin6->sin6_port, sizeof(in_port_t));
         }
     }
     return bnodes;
@@ -944,7 +1044,7 @@ NetworkEngine::bufferNodes(sa_family_t af, const InfoHash& id, std::vector<std::
 
 std::pair<Blob, Blob>
 NetworkEngine::bufferNodes(sa_family_t af, const InfoHash& id, want_t want,
-        std::vector<std::shared_ptr<Node>>& nodes4, std::vector<std::shared_ptr<Node>>& nodes6)
+        std::vector<Sp<Node>>& nodes4, std::vector<Sp<Node>>& nodes6)
 {
     if (want < 0)
         want = af == AF_INET ? WANT4 : WANT6;
@@ -960,19 +1060,47 @@ NetworkEngine::bufferNodes(sa_family_t af, const InfoHash& id, want_t want,
     return {std::move(bnodes4), std::move(bnodes6)};
 }
 
-std::shared_ptr<Request>
-NetworkEngine::sendListen(std::shared_ptr<Node> n, const InfoHash& infohash, const Query& query, const Blob& token,
-        RequestCb on_done, RequestExpiredCb on_expired) {
-    auto tid = TransId {TransPrefix::LISTEN, getNewTid()};
+Sp<Request>
+NetworkEngine::sendListen(Sp<Node> n,
+        const InfoHash& hash,
+        const Query& query,
+        const Blob& token,
+        Sp<Request> previous,
+        RequestCb&& on_done,
+        RequestExpiredCb&& on_expired,
+        SocketCb&& socket_cb)
+{
+    Sp<Socket> socket;
+    auto tid = TransId { TransPrefix::LISTEN, previous ? previous->tid.getTid() : getNewTid() };
+    if (previous and previous->node == n) {
+        socket = previous->socket;
+    } else {
+        if (previous)
+            DHT_LOG.e(hash, "[node %s] trying refresh listen contract with wrong node", previous->node->toString().c_str());
+        /* TODO: Manually creating a socket for "listen" backward compatibility.
+         * As soon as this is not an issue anymore, switch to using
+         * ::createSocket function! */
+        auto sid = TransId {TransPrefix::GET_VALUES, tid.getTid()};
+        socket = std::make_shared<Socket>(n, sid, socket_cb);
+        opened_sockets.emplace(sid, socket);
+    }
+
+    if (not socket) {
+        DHT_LOG.e(hash, "[node %s] unable to get a valid socket for listen. Aborting listen", n->toString().c_str());
+        return {};
+    }
+
     msgpack::sbuffer buffer;
     msgpack::packer<msgpack::sbuffer> pk(&buffer);
     pk.pack_map(5+(network?1:0));
 
     auto has_query = query.where.getFilter() or not query.select.getSelection().empty();
-    pk.pack(std::string("a")); pk.pack_map(3 + has_query);
+    pk.pack(std::string("a")); pk.pack_map(4 + has_query);
       pk.pack(std::string("id"));    pk.pack(myid);
-      pk.pack(std::string("h"));     pk.pack(infohash);
+      pk.pack(std::string("h"));     pk.pack(hash);
       pk.pack(std::string("token")); packToken(pk, token);
+      pk.pack(std::string("sid"));  pk.pack_bin(socket->id.size());
+                                     pk.pack_bin_body((const char*)socket->id.data(), socket->id.size());
       if (has_query) {
           pk.pack(std::string("q")); pk.pack(query);
       }
@@ -986,8 +1114,8 @@ NetworkEngine::sendListen(std::shared_ptr<Node> n, const InfoHash& infohash, con
         pk.pack(std::string("n")); pk.pack(network);
     }
 
-    Blob b {buffer.data(), buffer.data() + buffer.size()};
-    std::shared_ptr<Request> req(new Request {tid.getTid(), n, std::move(b),
+    auto req = std::make_shared<Request>(tid, n,
+        Blob(buffer.data(), buffer.data() + buffer.size()),
         [=](const Request& req_status, ParsedMessage&& msg) { /* on done */
             if (on_done)
                 on_done(req_status, {std::forward<ParsedMessage>(msg)});
@@ -996,8 +1124,8 @@ NetworkEngine::sendListen(std::shared_ptr<Node> n, const InfoHash& infohash, con
             if (on_expired)
                 on_expired(req_status, done);
         },
-        true
-    });
+        socket
+    );
     sendRequest(req);
     ++out_stats.listen;
     return req;
@@ -1024,14 +1152,14 @@ NetworkEngine::sendListenConfirmation(const SockAddr& addr, TransId tid) {
     send(buffer.data(), buffer.size(), 0, addr);
 }
 
-std::shared_ptr<Request>
-NetworkEngine::sendAnnounceValue(std::shared_ptr<Node> n,
+Sp<Request>
+NetworkEngine::sendAnnounceValue(Sp<Node> n,
         const InfoHash& infohash,
-        const std::shared_ptr<Value>& value,
+        const Sp<Value>& value,
         time_point created,
         const Blob& token,
-        RequestCb on_done,
-        RequestExpiredCb on_expired)
+        RequestCb&& on_done,
+        RequestExpiredCb&& on_expired)
 {
     auto tid = TransId {TransPrefix::ANNOUNCE_VALUES, getNewTid()};
     msgpack::sbuffer buffer;
@@ -1057,8 +1185,8 @@ NetworkEngine::sendAnnounceValue(std::shared_ptr<Node> n,
         pk.pack(std::string("n")); pk.pack(network);
     }
 
-    Blob b {buffer.data(), buffer.data() + buffer.size()};
-    std::shared_ptr<Request> req(new Request {tid.getTid(), n, std::move(b),
+    auto req = std::make_shared<Request>(tid, n,
+        Blob(buffer.data(), buffer.data() + buffer.size()),
         [=](const Request& req_status, ParsedMessage&& msg) { /* on done */
             if (msg.value_id == Value::INVALID_ID) {
                 DHT_LOG.d(infohash, "Unknown search or announce!");
@@ -1075,7 +1203,7 @@ NetworkEngine::sendAnnounceValue(std::shared_ptr<Node> n,
                 on_expired(req_status, done);
             }
         }
-    });
+    );
     sendRequest(req);
     if (not v.empty())
         sendValueParts(tid, v, n->addr);
@@ -1083,13 +1211,13 @@ NetworkEngine::sendAnnounceValue(std::shared_ptr<Node> n,
     return req;
 }
 
-std::shared_ptr<Request>
-NetworkEngine::sendRefreshValue(std::shared_ptr<Node> n,
+Sp<Request>
+NetworkEngine::sendRefreshValue(Sp<Node> n,
                 const InfoHash& infohash,
                 const Value::Id& vid,
                 const Blob& token,
-                RequestCb on_done,
-                RequestExpiredCb on_expired)
+                RequestCb&& on_done,
+                RequestExpiredCb&& on_expired)
 {
     auto tid = TransId {TransPrefix::REFRESH, getNewTid()};
     msgpack::sbuffer buffer;
@@ -1111,8 +1239,8 @@ NetworkEngine::sendRefreshValue(std::shared_ptr<Node> n,
         pk.pack(std::string("n")); pk.pack(network);
     }
 
-    Blob b {buffer.data(), buffer.data() + buffer.size()};
-    std::shared_ptr<Request> req(new Request {tid.getTid(), n, std::move(b),
+    auto req = std::make_shared<Request>(tid, n,
+        Blob(buffer.data(), buffer.data() + buffer.size()),
         [=](const Request& req_status, ParsedMessage&& msg) { /* on done */
             if (msg.value_id == Value::INVALID_ID) {
                 DHT_LOG.d(infohash, "Unknown search or announce!");
@@ -1129,11 +1257,10 @@ NetworkEngine::sendRefreshValue(std::shared_ptr<Node> n,
                 on_expired(req_status, done);
             }
         }
-    });
+    );
     sendRequest(req);
     ++out_stats.refresh;
     return req;
-
 }
 
 void
@@ -1163,7 +1290,8 @@ NetworkEngine::sendError(const SockAddr& addr,
         TransId tid,
         uint16_t code,
         const std::string& message,
-        bool include_id) {
+        bool include_id)
+{
     msgpack::sbuffer buffer;
     msgpack::packer<msgpack::sbuffer> pk(&buffer);
     pk.pack_map(4 + (include_id?1:0));
@@ -1260,6 +1388,9 @@ ParsedMessage::msgpack_unpack(msgpack::object msg)
         error_code = e->via.array.ptr[0].as<uint16_t>();
     }
 
+    if (auto t = findMapValue(req, "sid"))
+        socket_id = {t->as<std::array<char, 4>>()};
+
     if (auto rid = findMapValue(req, "id"))
         id = {*rid};
 
@@ -1369,9 +1500,9 @@ ParsedMessage::msgpack_unpack(msgpack::object msg)
 void
 NetworkEngine::maintainRxBuffer(const TransId& tid)
 {
-    const auto& now = scheduler.time();
     auto msg = partial_messages.find(tid);
     if (msg != partial_messages.end()) {
+        const auto& now = scheduler.time();
         if (msg->second.start + RX_MAX_PACKET_TIME < now
          || msg->second.last_part + RX_TIMEOUT < now) {
             DHT_LOG.w("Dropping expired partial message from %s", msg->second.from.toString().c_str());
@@ -1419,4 +1550,5 @@ ParsedMessage::complete()
 }
 
 
-}
+} /* namespace net  */
+} /* namespace dht */
