@@ -1,7 +1,8 @@
 /*
- *  Copyright (C) 2014-2016 Savoir-faire Linux Inc.
+ *  Copyright (C) 2014-2017 Savoir-faire Linux Inc.
  *
  *  Author: Adrien Béraud <adrien.beraud@savoirfairelinux.com>
+ *          Simon Désaulniers <simon.desaulniers@savoirfairelinux.com>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -14,8 +15,7 @@
  *  GNU General Public License for more details.
  *
  *  You should have received a copy of the GNU General Public License
- *  along with this program; if not, write to the Free Software
- *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA.
+ *  along with this program. If not, see <https://www.gnu.org/licenses/>.
  */
 
 #include "tools_common.h"
@@ -101,12 +101,10 @@ void cmd_loop(std::shared_ptr<DhtRunner>& dht, dht_params& params)
             continue;
         } else if (op == "ll") {
             print_node_info(dht, params);
-            unsigned good4, dubious4, cached4, incoming4;
-            unsigned good6, dubious6, cached6, incoming6;
-            dht->getNodesStats(AF_INET, &good4, &dubious4, &cached4, &incoming4);
-            dht->getNodesStats(AF_INET6, &good6, &dubious6, &cached6, &incoming6);
-            std::cout << "IPv4 nodes : " << good4 << " good, " << dubious4 << " dubious, " << incoming4 << " incoming." << std::endl;
-            std::cout << "IPv6 nodes : " << good6 << " good, " << dubious6 << " dubious, " << incoming6 << " incoming." << std::endl;
+            std::cout << "IPv4 stats:" << std::endl;
+            std::cout << dht->getNodesStats(AF_INET).toString() << std::endl;
+            std::cout << "IPv6 stats:" << std::endl;
+            std::cout << dht->getNodesStats(AF_INET6).toString() << std::endl;
             continue;
         } else if (op == "lr") {
             std::cout << "IPv4 routing table:" << std::endl;
@@ -361,6 +359,8 @@ main(int argc, char **argv)
         }
         if (params.daemonize) {
             daemonize();
+        } else if (params.service) {
+            setupSignals();
         }
 
         dht::crypto::Identity crt {};
@@ -372,7 +372,9 @@ main(int argc, char **argv)
         dht->run(params.port, crt, true, params.network);
 
         if (params.log) {
-            if (not params.logfile.empty())
+            if (params.syslog or (params.daemonize and params.logfile.empty()))
+                log::enableSyslog(*dht, "dhtnode");
+            else if (not params.logfile.empty())
                 log::enableFileLogging(*dht, params.logfile);
             else
                 log::enableLogging(*dht);
@@ -383,9 +385,8 @@ main(int argc, char **argv)
             dht->bootstrap(params.bootstrap.first.c_str(), params.bootstrap.second.c_str());
         }
 
-        if (params.daemonize) {
-            while (true)
-                std::this_thread::sleep_for(std::chrono::seconds(30));
+        if (params.daemonize or params.service) {
+            while (runner.wait());
         } else {
             cmd_loop(dht, params);
         }
